import asyncio
import json
import sys

import typer
from stlog import setup

from smartjob.app.executor import ExecutorService
from smartjob.app.input import GcsInput, Input, LocalPathInput
from smartjob.app.job import SmartJob
from smartjob.infra.controllers.lib import get_executor_service_singleton


def get_job_service(ctx: typer.Context, **kwargs) -> ExecutorService:
    return get_executor_service_singleton(
        namespace=ctx.obj.namespace,
        project=ctx.obj.project,
        region=ctx.obj.region,
        **kwargs,
    )


def init_stlog(ctx: typer.Context):
    setup(level=ctx.obj.log_level)


NameArgument = typer.Argument(help="Name of the job")
DockerImageArgument = typer.Option("", help="Docker image to use")
OverrideCommandArgument = typer.Option(
    "",
    help="Override docker image command and arguments",
)
AddEnvArgument = typer.Option(
    [],
    help="Add a new env var in the container environment (format: key=value, can be used multiple times)",
)
StagingBucketArgument = typer.Option(
    None,
    envvar="SMARTJOB_STAGING_BUCKET",
    help="staging bucket (starting with gs://) for loading python_script_path",
)
InputBucketBasePathArgument = typer.Option(
    None, envvar="SMARTJOB_INPUT_BUCKET_BASE_PATH"
)
OutputBucketBasePathArgument: str = typer.Option(
    None, envvar="SMARTJOB_OUTPUT_BUCKET_BASE_PATH"
)
InputBucketPathArgument = typer.Option("")
OutputBucketPathArgument = typer.Option("")
PythonScriptPathArgument = typer.Option(
    "", help="local path to python script to execute in the container"
)
WaitArgument = typer.Option(True, help="Wait for the job to finish")
LocalPathInputArgument = typer.Option(
    [],
    help="Local path input (format: filename=local_path, can be used multiple times)",
)
GcsPathInputArguments = typer.Option(
    [],
    help="GCS path input (format: filename=full_source_gcs_path, can be used multiple times)",
)


def add_env_argument_to_dict(add_env: list[str]) -> dict[str, str]:
    add_envs: dict[str, str] = {}
    for ae in add_env:
        if "=" not in ae:
            raise ValueError(f"Invalid env var format: {ae} => must be key=value")
        key, value = ae.split("=", 1)
        add_envs[key] = value
    return add_envs


async def fire_and_forget_job(
    service: ExecutorService, job: SmartJob, inputs: list[Input]
):
    future = await service.schedule(job, inputs=inputs)
    print("SCHEDULED")
    print("Id:      %s" % future.execution_id)
    print("Logs:    %s" % future.log_url)


def cli_process(
    service: ExecutorService, job: SmartJob, wait: bool, inputs: list[Input]
):
    if wait:
        result = service.sync_run(job, inputs=inputs)
        return_code = 0
        if result:
            print("SUCCESS in %i seconds" % (result.duration_seconds or -1))
        else:
            print("FAILED in %i seconds" % (result.duration_seconds or -1))
            return_code = 2
        print("Id:      %s" % result.execution_id)
        print("Logs:    %s" % result.log_url)
        if result.json_output is not None:
            print("JsonOutput:")
            print(json.dumps(result.json_output, indent=4))
        sys.exit(return_code)
    else:
        asyncio.run(fire_and_forget_job(service, job, inputs))


def local_path_input_to_list(local_path_input: list[str]) -> list[Input]:
    res: list[Input] = []
    for lpi in local_path_input:
        if "=" not in lpi:
            raise ValueError(
                f"Invalid input format: {lpi} => must be filename=local_path"
            )
        filename, local_path = lpi.split("=", 1)
        res.append(
            LocalPathInput(
                filename=filename,
                local_path=local_path,
            )
        )
    return res


def gcs_input_to_list(gcs_input: list[str]) -> list[Input]:
    res: list[Input] = []
    for lpi in gcs_input:
        if "=" not in lpi:
            raise ValueError(
                f"Invalid input format: {lpi} => must be filename=full_gcs_path"
            )
        filename, gcs_path = lpi.split("=", 1)
        if not gcs_path.startswith("gs://"):
            raise ValueError(f"Invalid GCS path: {gcs_path} => must start with gs://")
        if "/" not in gcs_path[5:]:
            raise ValueError(
                f"Invalid GCS path: {gcs_path} => must have a / after gs://"
            )
        res.append(
            GcsInput(
                filename=filename,
                gcs_path=gcs_path,
            )
        )
    return res



import shlex

import typer

from smartjob.app.job import CloudRunSmartJob
from smartjob.infra.controllers.cli.utils import (
    AddEnvArgument,
    DockerImageArgument,
    GcsPathInputArguments,
    LocalPathInputArgument,
    NameArgument,
    OverrideCommandArgument,
    PythonScriptPathArgument,
    StagingBucketArgument,
    WaitArgument,
    add_env_argument_to_dict,
    cli_process,
    gcs_input_to_list,
    get_job_service,
    init_stlog,
    local_path_input_to_list,
)

cli = typer.Typer()


@cli.command()
def run(
    ctx: typer.Context,
    name: str = NameArgument,
    docker_image: str = DockerImageArgument,
    override_command_and_args: str = OverrideCommandArgument,
    add_env: list[str] = AddEnvArgument,
    staging_bucket: str = StagingBucketArgument,
    python_script_path: str = PythonScriptPathArgument,
    wait: bool = WaitArgument,
    cpu: float = typer.Option(1.0, help="Number of CPUs"),
    memory_gb: float = typer.Option(0.5, help="Memory in Gb"),
    local_path_input: list[str] = LocalPathInputArgument,
    gcs_input: list[str] = GcsPathInputArguments,
):
    init_stlog(ctx)
    overriden_args = shlex.split(override_command_and_args)
    add_envs = add_env_argument_to_dict(add_env)
    inputs = local_path_input_to_list(local_path_input) + gcs_input_to_list(gcs_input)
    service = get_job_service(ctx)
    job = CloudRunSmartJob(
        name=name,
        docker_image=docker_image,
        overridden_args=overriden_args,
        add_envs=add_envs,
        staging_bucket=staging_bucket,
        python_script_path=python_script_path,
        cpu=cpu,
        memory_gb=memory_gb,
    )
    cli_process(service, job, wait, inputs)



import asyncio
import json

from smartjob.app.storage import StoragePort


class DummyStorageAdapter(StoragePort):
    def __init__(self, sleep: float = 1.0):
        self._sleep = sleep

    async def sleep(self):
        await asyncio.sleep(self._sleep)

    async def upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ) -> str:
        await self.sleep()
        return f"{destination_bucket}/{destination_path}"

    async def copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ) -> str:
        await self.sleep()
        return f"{destination_bucket}/{destination_path}"

    async def download(
        self,
        source_bucket: str,
        source_path: str,
    ) -> bytes:
        await self.sleep()
        return json.dumps({"dummy": "result"}).encode()



import asyncio
import concurrent.futures
import logging
import shlex
import sys
import typing
from threading import Lock

from google.cloud import aiplatform
from google.cloud.aiplatform_v1.types import custom_job as custom_job_v1
from google.cloud.aiplatform_v1.types import job_state as gca_job_state
from google.cloud.aiplatform_v1.types.env_var import EnvVar
from google.cloud.aiplatform_v1.types.machine_resources import DiskSpec, MachineSpec
from stlog import getLogger

from smartjob.app.executor import ExecutionResultFuture, ExecutorPort
from smartjob.app.job import (
    Execution,
    ExecutionResult,
    VertexSmartJob,
)

aiplatform_mutex = Lock()
aiplatform_initialized: bool = False
logger = getLogger("smartjob.executor.vertex")


def monkey_patch_aiplatform_jobs_logger():
    from google.cloud.aiplatform.jobs import _LOGGER

    for handler in _LOGGER.handlers:
        _LOGGER.removeHandler(handler)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.WARNING)
    _LOGGER.addHandler(handler)
    logging.getLogger("google.cloud.aiplatform.jobs").setLevel(logging.WARNING)


def init_aiplatform():
    global aiplatform_initialized
    with aiplatform_mutex:
        if not aiplatform_initialized:
            aiplatform.init()
            monkey_patch_aiplatform_jobs_logger()
            aiplatform_initialized = True


class VertexCustomJob(aiplatform.CustomJob):
    @property
    def id(self) -> str:
        return self.resource_name.split("/")[-1]

    def get_log_url(self, project) -> str:
        return f"https://console.cloud.google.com/logs/query;query=resource.labels.job_id%3D%22{self.id}%22?project={project}"

    @property
    def success(self) -> bool:
        return self.state == gca_job_state.JobState.JOB_STATE_SUCCEEDED


class VertexExecutionResultFuture(ExecutionResultFuture):
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        return future.result()


class VertexExecutor(ExecutorPort):
    def __init__(self, max_workers: int = 10):
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)

    def init_aiplatform_if_needed(self):
        init_aiplatform()

    def sync_run(self, execution: Execution) -> ExecutionResult:
        self.init_aiplatform_if_needed()
        job = typing.cast(VertexSmartJob, execution.job)
        customJob = VertexCustomJob(
            display_name=f"{job.namespace}-{job.name}-{execution.id}",
            project=job.project,
            location=job.region,
            staging_bucket=job.staging_bucket,
            worker_pool_specs=[
                custom_job_v1.WorkerPoolSpec(
                    replica_count=1,
                    container_spec=custom_job_v1.ContainerSpec(
                        image_uri=job.docker_image,
                        args=execution.overridden_args,
                        env=[
                            EnvVar(name=k, value=v)
                            for k, v in execution.add_envs.items()
                        ],
                    ),
                    machine_spec=MachineSpec(
                        machine_type=job.machine_type,
                        accelerator_type=job.accelerator_type,
                        accelerator_count=job.accelerator_count,
                    ),
                    disk_spec=DiskSpec(
                        boot_disk_type=job.boot_disk_type,
                        boot_disk_size_gb=job.boot_disk_size_gb,
                    ),
                )
            ],
        )
        logger.info(
            "Let's trigger a new execution of Vertex Job: %s...",
            f"{job.namespace}-{job.name}-{execution.id}",
            docker_image=job.docker_image,
            overridden_args=shlex.join(execution.overridden_args),
            add_envs=", ".join([f"{x}={y}" for x, y in execution.add_envs.items()]),
        )
        try:
            customJob.submit(
                disable_retries=True,
                timeout=job.timeout_seconds,
                service_account=job.service_account,
            )
            execution.log_url = customJob.get_log_url(job.project)
            customJob._block_until_complete()
        except RuntimeError:
            pass
        return ExecutionResult.from_execution(execution, customJob.success)

    # TODO: we can probably do this in a better way
    async def _wait(self, future: asyncio.Future[ExecutionResult]) -> ExecutionResult:
        while True:
            if future.done():
                return future.result()
            await asyncio.sleep(1)

    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        loop = asyncio.get_event_loop()
        future = loop.run_in_executor(self.executor, self.sync_run, execution)
        return VertexExecutionResultFuture(self._wait(future), execution=execution)



import asyncio

from smartjob.app.executor import ExecutionResultFuture, ExecutorPort
from smartjob.app.job import (
    Execution,
    ExecutionResult,
)


class DummyExecutionResultFuture(ExecutionResultFuture):
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        return future.result()


class DummyExecutor(ExecutorPort):
    def __init__(self, sleep: float = 1.0):
        self._sleep = sleep

    async def work(self, execution: Execution) -> ExecutionResult:
        await asyncio.sleep(self._sleep)
        return ExecutionResult.from_execution(execution, True)

    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        execution.log_url = "https://example.com/log"
        return DummyExecutionResultFuture(self.work(execution), execution)



from abc import ABC, abstractmethod


class StoragePort(ABC):
    @abstractmethod
    async def download(
        self,
        source_bucket: str,
        source_path: str,
    ) -> bytes:
        pass

    @abstractmethod
    async def upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ) -> str:
        pass

    @abstractmethod
    async def copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ) -> str:
        pass



import os
from typing import Any

import stlog

from smartjob.app.executor import ExecutorService
from smartjob.infra.adapters.executor.cloudrun import CloudRunExecutor
from smartjob.infra.adapters.executor.dummy import DummyExecutor
from smartjob.infra.adapters.executor.vertex import VertexExecutor
from smartjob.infra.adapters.storage.dummy import DummyStorageAdapter
from smartjob.infra.adapters.storage.gcs import GcsStorageAdapter

# Default max workers for vertex executor and for file uploader (cloud run executor is not multi-threaded)
DEFAULT_MAX_WORKERS = 10
__singleton: ExecutorService | None = None


def get_executor_service_singleton(
    max_workers: int = DEFAULT_MAX_WORKERS,
    namespace: str | None = None,
    project: str | None = None,
    region: str | None = None,
    staging_bucket: str | None = None,
    docker_image: str | None = None,
) -> ExecutorService:
    """Return a singleton instance of ExecutorService (initialized on first call with given arguments).

    Most of the arguments are optional and can also be set via environment variables or at SmartJob level.

    See ExecutorService or SmartJob for more details on the arguments.

    Note: if SMARTJOB_USE_DUMMY_EXECUTOR is set to 'true', a dummy executor will be returned.
          (only useful for testing or debugging)

    Args:
        max_workers: Maximum number of workers for the vertex executor and for the file uploader.
        namespace: Default namespace to use.
        project: Default project to use.
        region: Default region to use.
        staging_bucket: Default staging bucket to use (for input/output/local script uploading)
        docker_image: Default docker image to use.

    Returns:
        Instance of ExecutorService.

    """
    global __singleton
    kwargs: dict[str, Any] = {}
    if docker_image:
        kwargs["docker_image"] = docker_image
    if staging_bucket:
        kwargs["staging_bucket"] = staging_bucket
    if region:
        kwargs["region"] = region
    if project:
        kwargs["project"] = project
    if namespace:
        kwargs["namespace"] = namespace
    if os.environ.get("SMARTJOB_USE_DUMMY_EXECUTOR", "false").lower() == "true":
        stlog.getLogger("smartjob").warning(
            "Using dummy executor because SMARTJOB_USE_DUMMY_EXECUTOR is set to true"
        )
        return ExecutorService(
            cloudrun_executor_adapter=DummyExecutor(),
            vertex_executor_adapter=DummyExecutor(),
            storage_adapter=DummyStorageAdapter(),
            **kwargs,
        )
    if __singleton is None:
        __singleton = ExecutorService(
            cloudrun_executor_adapter=CloudRunExecutor(),
            vertex_executor_adapter=VertexExecutor(max_workers=max_workers),
            storage_adapter=GcsStorageAdapter(max_workers=max_workers),
            **kwargs,
        )
    return __singleton



import datetime
import hashlib
import json
import uuid
from dataclasses import dataclass, field

from smartjob.app.exception import SmartJobException

# DEFAULT_NAMESPACE is the default namespace for jobs.
DEFAULT_NAMESPACE = "default"

# DEFAULT_TIMEOUT_SECONDS is the default timeout in seconds for jobs.
DEFAULT_TIMEOUT_SECONDS = 3600


def _hex_hash(*args: str) -> str:
    return hashlib.sha1("\n".join(args).encode()).hexdigest()


def _unique_id() -> str:
    return str(uuid.uuid4()).replace("-", "")


@dataclass
class SmartJob:
    """Base class for smart jobs."""

    name: str
    """Name of the job."""

    namespace: str = DEFAULT_NAMESPACE
    """Namespace of the job."""

    project: str = ""
    """GCP project to use for executing the job.

    If not set here, the value will be forced by the ExecutorService object.
    """

    region: str = ""
    """GCP region where to execute the job.

    If not set here, the value will be forced by the ExecutorService object.
    """

    docker_image: str = ""
    """Docker image to use for the job.

    Example: `docker.io/python:3.12`.
    """

    add_envs: dict[str, str] = field(default_factory=dict)
    """Environnement variables to add in the container.

    Note: `INPUT_PATH` and `OUTPUT_PATH` will be automatically injected (if relevant).
    """

    overridden_args: list[str] = field(default_factory=list)
    """Container arguments (including command) to use.

    If not set, the default container image arguments will be used.

    """

    python_script_path: str = ""
    """Local path to a python script to execute in the container."""

    staging_bucket: str = ""
    """Staging bucket (for input, output and/or loading python_script_path."""

    timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS
    """"Timeout in seconds for the job execution."""

    service_account: str | None = None
    """Service account (email) to use for the job execution."""

    def _assert_is_ready(self):
        if not self.name:
            raise SmartJobException("name is required for job")
        if not self.project:
            raise SmartJobException("project is required")
        if not self.region:
            raise SmartJobException("region is required")
        if not self.docker_image:
            raise SmartJobException("docker_image is required")
        if not self.staging_bucket:
            raise SmartJobException("staging_bucket is required")
        if not self.staging_bucket.startswith("gs://"):
            raise SmartJobException("staging_bucket must start with gs://")
        if "/" in self.staging_bucket[5:]:
            raise SmartJobException(
                "staging_bucket name must not contain / (except for the the gs:// prefix)"
            )

    @property
    def staging_bucket_name(self) -> str:
        """Return the name of the staging bucket without the gs:// prefix.

        If it does not start with gs://, it returns an empty string.

        """
        if not self.staging_bucket.startswith("gs://"):
            return ""
        return self.staging_bucket[5:]  # let's remove gs:// prefix

    @property
    def _staging_mount_point(self) -> str:
        """Return the mount point for the staging bucket."""
        raise NotImplementedError("mount_point must be implemented in subclasses")

    @property
    def full_name(self) -> str:
        """Return the full name of the job (namespace + name)."""
        return f"{self.namespace}-{self.name}"


@dataclass
class CloudRunSmartJob(SmartJob):
    cpu: float = 1
    """Number of requested CPUs."""

    memory_gb: float = 0.5
    """Memory in Gb."""

    @property
    def _job_hash(self) -> str:
        """Return a short stable job_hash.

        If the job_hash is the same, we don't have to create a new CloudRunJob instance
        (but only a new job execution).

        """
        args: list[str] = [self.name, self.project, self.region, self.docker_image]
        args += [self.namespace, self.staging_bucket, self.service_account or ""]
        return _hex_hash(*args)[:10]

    @property
    def _parent_name(self) -> str:
        return f"projects/{self.project}/locations/{self.region}"

    @property
    def _cloud_run_job_name(self) -> str:
        """Return the cloud run job name."""
        return f"{self.namespace}-{self.name}-{self._job_hash}"

    @property
    def _full_cloud_run_job_name(self) -> str:
        """Return the full cloud run job name."""
        return f"{self._parent_name}/jobs/{self._cloud_run_job_name}"

    @property
    def _cpuLimit(self) -> str:
        return "%f" % self.cpu

    @property
    def _memoryLimit(self) -> str:
        return "%f" % self.memory_gb + "Gi"

    @property
    def _staging_mount_point(self) -> str:
        """Return the mount point for the staging bucket."""
        return "/staging"


@dataclass
class VertexSmartJob(SmartJob):
    machine_type: str = "n1-standard-4"
    """Machine type."""

    accelerator_type: str = "ACCELERATOR_TYPE_UNSPECIFIED"
    """Accelerator type."""

    accelerator_count: int = 0
    """Number of accelerators."""

    boot_disk_type: str = "pd-ssd"
    """Boot disk type."""

    boot_disk_size_gb: int = 100
    """Boot disk size in Gb."""

    @property
    def _staging_mount_point(self) -> str:
        """Return the mount point for the staging bucket."""
        return f"/gcs/{self.staging_bucket_name}"

    def _assert_is_ready(self):
        """Test if the job is ready to be run.

        It raises an exception if the job is not ready to be run.
        If it's ok, it does nothing.

        Raises:
            SmartJobException: If the job is not ready to be run.

        """
        super()._assert_is_ready()
        if not self.staging_bucket:
            raise SmartJobException("staging_bucket is required for vertex jobs")


@dataclass
class Execution:
    job: SmartJob
    id: str = field(default_factory=_unique_id)
    created: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now(tz=datetime.timezone.utc),
        init=False,
    )
    log_url: str | None = field(default=None, init=False)
    cancelled: bool = field(default=False, init=False)
    add_envs: dict[str, str] = field(default_factory=dict)
    overridden_args: list[str] = field(default_factory=list)

    @property
    def base_dir(self) -> str:
        dat = self.created.strftime("%Y-%m-%d")
        tim = self.created.strftime("%H%M")
        return f"{self.job.full_name}/{dat}/{tim}/{self.id}"

    @property
    def _input_path(self) -> str:
        return f"{self.base_dir}/input"

    @property
    def _output_path(self) -> str:
        return f"{self.base_dir}/output"

    @property
    def input_path(self) -> str:
        return f"{self.job._staging_mount_point}/{self._input_path}"

    @property
    def output_path(self) -> str:
        return f"{self.job._staging_mount_point}/{self._output_path}"

    @property
    def full_input_path(self) -> str:
        return f"{self.job.staging_bucket}/{self._input_path}"

    @property
    def full_output_path(self) -> str:
        return f"{self.job.staging_bucket}/{self._output_path}"


@dataclass
class ExecutionResult:
    """ExecutionResult is the (final) result of a job execution.

    Attributes:
        success: Whether the job has succeeded or not.
        created: The datetime when the job has started.
        stopped: The datetime when the job has stopped.
        execution_id: The execution id of the job.
        job_name: The name of the job.
        job_namespace: The namespace of the job.
        log_url: The execution log url.
        full_input_path: The full input path (starting with gs://).
        full_output_path: The full output path (starting with gs//).
        json_output: if the job has created a json file named smartjob.json in the output directory, it will be stored/decoded here.

    """

    success: bool
    created: datetime.datetime
    stopped: datetime.datetime
    execution_id: str
    job_name: str
    job_namespace: str
    log_url: str
    full_input_path: str
    full_output_path: str
    json_output: dict | list | str | float | int | bool | None = None

    def __bool__(self) -> bool:
        return self.success or False

    def __str__(self) -> str:
        if self.success:
            state = "SUCCESS"
        else:
            state = "FAILURE"
        if self.json_output is not None:
            json_output = json.dumps(self.json_output, indent=4)
        else:
            json_output = "None"
        res = f"""ExecutionResult(
    job_name={self.job_name}, job_namespace={self.job_namespace},
    execution_id={self.execution_id},
    state={state}, duration_seconds={self.duration_seconds},
    log_url={self.log_url},
    json_output={json_output}
)"""
        return res

    @property
    def duration_seconds(self) -> int:
        """The duration of the job in seconds."""
        return (self.stopped - self.created).seconds

    @classmethod
    def from_execution(
        cls,
        execution: Execution,
        success: bool,
    ) -> "ExecutionResult":
        """Create a ExecutionResult from a SmartJobExecution."""
        if not execution.log_url:
            raise SmartJobException("log_url is required")
        return cls(
            created=execution.created,
            execution_id=execution.id,
            job_name=execution.job.name,
            job_namespace=execution.job.namespace,
            log_url=execution.log_url,
            success=success,
            stopped=datetime.datetime.now(tz=datetime.timezone.utc),
            full_input_path=execution.full_input_path,
            full_output_path=execution.full_output_path,
        )



from smartjob.app.exception import SmartJobException
from smartjob.app.executor import ExecutionResultFuture, ExecutorService
from smartjob.app.input import BytesInput, GcsInput, Input, JsonInput, LocalPathInput
from smartjob.app.job import (
    CloudRunSmartJob,
    ExecutionResult,
    SmartJob,
    VertexSmartJob,
)
from smartjob.infra.controllers.lib import get_executor_service_singleton

__all__ = [
    "get_executor_service_singleton",
    "SmartJobException",
    "CloudRunSmartJob",
    "SmartJob",
    "ExecutorService",
    "ExecutionResult",
    "ExecutionResultFuture",
    "VertexSmartJob",
    "Input",
    "JsonInput",
    "BytesInput",
    "GcsInput",
    "LocalPathInput",
]



import asyncio
import json
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Coroutine

from stlog import LogContext, getLogger

from smartjob.app.exception import SmartJobException
from smartjob.app.input import Input
from smartjob.app.job import (
    DEFAULT_NAMESPACE,
    CloudRunSmartJob,
    Execution,
    ExecutionResult,
    SmartJob,
    VertexSmartJob,
)
from smartjob.app.storage import StoragePort

logger = getLogger("smartjob.executor")


class ExecutionResultFuture(ABC):
    """Future-like object to get the result of a job execution.

    Attributes:
        job: The job that was scheduled.

    """

    def __init__(
        self,
        coroutine: Coroutine,
        execution: Execution,
    ):
        self._execution: Execution = execution
        self.__result: ExecutionResult | None = None
        self.__task = asyncio.create_task(coroutine)
        self.__task.add_done_callback(self._task_done)
        self._storage_adapter: StoragePort | None = None

    def _cancel(self):
        if not self.__result and not self._execution.cancelled:
            self._execution.cancelled = True
            self.__task.cancel()

    async def _download_output(self) -> dict | list | str | int | float | bool | None:
        if self._storage_adapter is None:
            return None
        logger.info("Downloading smartjob.json from output (if exists)...")
        raw = await self._storage_adapter.download(
            self._execution.job.staging_bucket_name,
            f"{self._execution._output_path}/smartjob.json",
        )
        if raw == b"":
            logger.debug("No smartjob.json found in output")
            return None
        try:
            res = json.loads(raw)
            logger.debug("smartjob.json downloaded/decoded")
            return res
        except Exception:
            logger.warning("smartjob.json is not a valid json")
            return None

    async def result(self) -> ExecutionResult:
        """Wait and return the result of the job execution.

        This is a coroutine. You have to wait for it to complete with (for example):
        `res = await future.result()`

        Returns:
            The result of the job execution.

        """
        if self.__result is not None:
            # when _task_done() called BEFORE result()
            self.__result.json_output = await self._download_output()
            return self.__result
        await self.__task  # blocking wait for result
        while True:
            # wait for _task_done() callback to be called
            if self.__result is not None:
                self.__result.json_output = await self._download_output()
                return self.__result
            await asyncio.sleep(0.1)

    def done(self) -> bool:
        """Return True if the job execution is done."""
        return self.__task.done()

    @property
    def execution_id(self) -> str:
        "The execution unique identifier."
        return self._execution.id

    @property
    def log_url(self) -> str:
        "The execution log url."
        if self._execution.log_url is None:
            raise SmartJobException("the log_url property is not set")
        return self._execution.log_url

    @abstractmethod
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        pass

    def _task_done(self, future: asyncio.Future):
        self.__result = self._get_result_from_future(future)
        if self._execution.cancelled:
            return
        if self.__result.success:
            logger.info(
                "Smartjob execution succeeded",
                duration_seconds=self.__result.duration_seconds,
            )
        else:
            logger.warn(
                "Smartjob execution failed",
                log_url=self.__result.log_url,
                duration_seconds=self.__result.duration_seconds,
            )


class ExecutorPort(ABC):
    @abstractmethod
    async def schedule(self, job: Execution) -> ExecutionResultFuture:
        pass


@dataclass
class ExecutorService:
    cloudrun_executor_adapter: ExecutorPort
    vertex_executor_adapter: ExecutorPort
    storage_adapter: StoragePort
    namespace: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_NAMESPACE", DEFAULT_NAMESPACE)
    )
    project: str = field(default_factory=lambda: os.environ.get("SMARTJOB_PROJECT", ""))
    region: str = field(default_factory=lambda: os.environ.get("SMARTJOB_REGION", ""))
    docker_image: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_DOCKER_IMAGE", "")
    )
    staging_bucket: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_STAGING_BUCKET", "")
    )

    def _update_job_with_default_parameters(self, job: SmartJob, execution_id: str):
        if not job.namespace:
            job.namespace = self.namespace
        if not job.project:
            job.project = self.project
        if not job.region:
            job.region = self.region
        if not job.docker_image:
            job.docker_image = self.docker_image
        if not job.staging_bucket:
            job.staging_bucket = self.staging_bucket

    def _update_execution_env(self, execution: Execution):
        execution.add_envs["INPUT_PATH"] = execution.input_path
        execution.add_envs["OUTPUT_PATH"] = execution.output_path
        execution.add_envs["EXECUTION_ID"] = execution.id

    async def _create_input_output_paths_if_needed(self, execution: Execution):
        job = execution.job
        coroutines: list[Coroutine] = []
        logger.info(
            "Creating input path gs://%s/%s/...",
            job.staging_bucket_name,
            execution._input_path,
        )
        coroutines.append(
            self.storage_adapter.upload(
                b"", job.staging_bucket_name, execution._input_path + "/"
            )
        )
        logger.info(
            "Creating output path gs://%s/%s/...",
            job.staging_bucket_name,
            execution._output_path,
        )
        coroutines.append(
            self.storage_adapter.upload(
                b"", job.staging_bucket_name, execution._output_path + "/"
            )
        )
        await asyncio.gather(*coroutines, return_exceptions=True)
        logger.debug("Done creating input/output paths")

    async def _upload_python_script_if_needed_and_update_overridden_args(
        self, execution: Execution
    ):
        job = execution.job
        if not job.python_script_path:
            return
        with open(job.python_script_path) as f:
            content = f.read()
        destination_path = f"{execution.base_dir}/script.py"
        logger.info(
            "Uploading python script (%s) to %s/%s...",
            job.python_script_path,
            job.staging_bucket,
            destination_path,
        )
        await self.storage_adapter.upload(
            content.encode("utf8"), job.staging_bucket_name, destination_path
        )
        logger.debug(
            "Done uploading python script (%s) to %s/%s",
            job.python_script_path,
            job.staging_bucket,
            destination_path,
        )
        execution.overridden_args = [
            "python",
            f"{job._staging_mount_point}/{destination_path}",
        ]

    async def _upload_inputs(self, execution: Execution, inputs: list[Input]):
        async def _upload_input(input: Input):
            path = f"gs://{execution.job.staging_bucket_name}/{execution._input_path}/{input.filename}"
            logger.info(f"Uploading input to {path}...")
            await input._create(
                execution.job.staging_bucket_name,
                execution._input_path,
                self.storage_adapter,
            )
            logger.debug(f"Done uploading input: {path}")

        await asyncio.gather(*[_upload_input(input) for input in inputs])

    async def schedule(
        self,
        job: SmartJob,
        add_envs: dict[str, str] | None = None,
        inputs: list[Input] | None = None,
    ) -> ExecutionResultFuture:
        """Schedule a job and return a kind of future (coroutine version).

        Arguments:
            job: The job to run.
            add_envs: Environment variables to add for this particular execution.
            inputs: Inputs to add for this particular execution.

        Returns:
            The result of the job execution as a kind of future.

        """
        execution = Execution(
            job,
            overridden_args=list(job.overridden_args),
            add_envs={**job.add_envs, **(add_envs or {})},
        )
        self._update_job_with_default_parameters(job, execution_id=execution.id)
        self._update_execution_env(execution)
        await self._upload_inputs(execution, inputs or [])
        job._assert_is_ready()
        LogContext.reset_context()
        LogContext.add(
            job_name=job.name,
            job_namespace=job.namespace,
            project=job.project,
            region=job.region,
            execution_id=execution.id,
        )
        logger.info("Starting a smartjob...")
        await self._create_input_output_paths_if_needed(execution)
        await self._upload_python_script_if_needed_and_update_overridden_args(execution)
        res: ExecutionResultFuture
        if isinstance(job, CloudRunSmartJob):
            res = await self.cloudrun_executor_adapter.schedule(execution)
        elif isinstance(job, VertexSmartJob):
            res = await self.vertex_executor_adapter.schedule(execution)
        else:
            raise SmartJobException("Unknown job type")
        res._storage_adapter = self.storage_adapter
        return res

    async def run(
        self,
        job: SmartJob,
        add_envs: dict[str, str] | None = None,
        inputs: list[Input] | None = None,
    ) -> ExecutionResult:
        """Schedule a job and wait for its completion (couroutine version).

        Arguments:
            job: The job to run.
            add_envs: Environment variables to add for this particular execution.
            inputs: Inputs to add for this particular execution.

        Returns:
            The result of the job execution.

        """
        future = await self.schedule(job, add_envs=add_envs, inputs=inputs)
        return await future.result()

    def sync_run(
        self,
        job: SmartJob,
        add_envs: dict[str, str] | None = None,
        inputs: list[Input] | None = None,
    ) -> ExecutionResult:
        """Schedule a job and wait for its completion (blocking version).

        Arguments:
            job: The job to run.
            add_envs: Environment variables to add for this particular execution.
            inputs: Inputs to add for this particular execution.

        Returns:
            The result of the job execution.
        """
        return asyncio.run(self.run(job, add_envs, inputs))



import asyncio
import shlex
from dataclasses import dataclass
from typing import cast

import google.api_core.exceptions as gac_exceptions
from google.cloud import run_v2
from stlog import getLogger

from smartjob.app.executor import ExecutionResultFuture, ExecutorPort
from smartjob.app.job import (
    CloudRunSmartJob,
    Execution,
    ExecutionResult,
)

logger = getLogger("smartjob.executor.cloudrun")


@dataclass(frozen=True)
class JobListCacheKey:
    project: str
    region: str


class CloudRunExecutionResultFuture(ExecutionResultFuture):
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        try:
            task_result = future.result()
            success = False
            castedResult = cast(run_v2.Execution, task_result)
            success = castedResult.succeeded_count > 0
            return ExecutionResult.from_execution(self._execution, success)
        except gac_exceptions.GoogleAPICallError:
            pass
        except asyncio.CancelledError:
            self._execution.cancelled = True
        return ExecutionResult.from_execution(self._execution, False)


class CloudRunExecutor(ExecutorPort):
    def __init__(self):
        self.job_list_cache: dict[JobListCacheKey, list[str]] | None = None
        self.client_cache: run_v2.JobsAsyncClient | None = None
        self.client_cache_lock = asyncio.Lock()

    @property
    def client(self) -> run_v2.JobsAsyncClient:
        # lazy-init to circumvent some 'attached to a different loop' issues
        if self.client_cache is None:
            self.client_cache = run_v2.JobsAsyncClient()
        return self.client_cache

    async def get_job_list_cache(self, project, region) -> list[str]:
        key = JobListCacheKey(project, region)
        if self.job_list_cache is not None:
            if key in self.job_list_cache:
                return self.job_list_cache[key]
        else:
            self.job_list_cache = {}
        logger.info(
            "Fetching Cloud Run Job list for %s/%s (to cache)...",
            project,
            region,
        )
        jobs = []
        req = run_v2.ListJobsRequest(parent=f"projects/{project}/locations/{region}")
        jobs_pager = await self.client.list_jobs(req)
        async for existing_job in jobs_pager:
            jobs.append(existing_job.name)
        self.job_list_cache[key] = jobs
        logger.info("%i existing Cloud Run Jobs found", len(jobs))
        return jobs

    def reset_job_list_cache(self, project, region):
        if self.job_list_cache is None:
            return
        key = JobListCacheKey(project, region)
        if key in self.job_list_cache:
            self.job_list_cache.pop(key)

    async def create_job_if_needed(self, job: CloudRunSmartJob):
        async with self.client_cache_lock:
            job_list_cache = await self.get_job_list_cache(job.project, job.region)
            if job._full_cloud_run_job_name in job_list_cache:
                # it already exists
                logger.debug(
                    "Cloud Run Job: %s already exists => great!",
                    job._full_cloud_run_job_name,
                )
                return
            # it does not exist => let's create it
            volumes: list[run_v2.Volume] = []
            volume_mounts: list[run_v2.VolumeMount] = []
            launch_stage = "GA"
            if job.staging_bucket_name:
                volumes.append(
                    run_v2.Volume(
                        name="staging",
                        gcs=run_v2.GCSVolumeSource(
                            bucket=job.staging_bucket_name, read_only=False
                        ),
                    )
                )
                volume_mounts.append(
                    run_v2.VolumeMount(name="staging", mount_path="/staging")
                )
                launch_stage = "BETA"
            request = run_v2.CreateJobRequest(
                parent=job._parent_name,
                job_id=job._cloud_run_job_name,
                job=run_v2.Job(
                    labels={
                        "smartjob": "true",
                        "smartjob.namespace": job.namespace,
                    },
                    launch_stage=launch_stage,
                    template=run_v2.ExecutionTemplate(
                        task_count=1,
                        template=run_v2.TaskTemplate(
                            execution_environment=run_v2.ExecutionEnvironment(
                                run_v2.ExecutionEnvironment.EXECUTION_ENVIRONMENT_GEN2
                            ),
                            containers=[
                                run_v2.Container(
                                    name="container-1",
                                    image=job.docker_image,
                                    resources=run_v2.ResourceRequirements(
                                        startup_cpu_boost=False,
                                        limits={
                                            "cpu": job._cpuLimit,
                                            "memory": job._memoryLimit,
                                        },
                                    ),
                                    volume_mounts=volume_mounts,
                                )
                            ],
                            volumes=volumes,
                            max_retries=1,
                            service_account=job.service_account,
                        ),
                    ),
                ),
            )
            logger.info(
                "Let's create a new Cloud Run Job: %s...", job._cloud_run_job_name
            )
            operation = await self.client.create_job(request=request)
            await operation.result()
            logger.debug("Done creating Cloud Run Job: %s", job._cloud_run_job_name)
            self.reset_job_list_cache(job.project, job.region)

    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        job = cast(CloudRunSmartJob, execution.job)
        await self.create_job_if_needed(job)
        request = run_v2.RunJobRequest(
            name=job._full_cloud_run_job_name,
            overrides=run_v2.RunJobRequest.Overrides(
                task_count=1,
                # timeout=job.timeout_seconds,
                container_overrides=[
                    run_v2.RunJobRequest.Overrides.ContainerOverride(
                        name="container-1",
                        args=execution.overridden_args,
                        env=[
                            run_v2.EnvVar(name=k, value=v)
                            for k, v in execution.add_envs.items()
                        ],
                    )
                ],
            ),
        )
        logger.info(
            "Let's trigger a new execution of Cloud Run Job: %s...",
            job._cloud_run_job_name,
            docker_image=job.docker_image,
            overridden_args=shlex.join(execution.overridden_args),
            add_envs=", ".join([f"{x}={y}" for x, y in execution.add_envs.items()]),
            timeout_s=job.timeout_seconds,
        )
        operation = await self.client.run_job(request=request)
        execution.log_url = operation.metadata.log_uri
        return CloudRunExecutionResultFuture(operation.result(), execution=execution)



from types import SimpleNamespace

import typer
import typer.core

from smartjob.infra.controllers.cli.cloudrun import cli as cloudrun_cli
from smartjob.infra.controllers.cli.vertex import cli as vertex_cli

# Disable rich usage for CLI doc generation
typer.core.rich = None  # type: ignore

cli = typer.Typer(add_completion=False)
cli.add_typer(cloudrun_cli, name="cloudrun")
cli.add_typer(vertex_cli, name="vertex")


@cli.callback()
def main(
    ctx: typer.Context,
    namespace: str = typer.Option(
        None, envvar="SMARTJOB_NAMESPACE", help="namespace (simple string)"
    ),
    project: str = typer.Option(None, envvar="SMARTJOB_PROJECT", help="project name"),
    region: str = typer.Option(None, envvar="SMARTJOB_REGION", help="region name"),
    log_level: str = typer.Option("INFO", envvar="LOG_LEVEL", help="log level"),
):
    ctx.obj = SimpleNamespace(
        namespace=namespace,
        region=region,
        project=project,
        log_level=log_level,
    )


if __name__ == "__main__":
    cli()
